#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
The Engine.
"""

import json
import logging
import os
import time
from contextlib import suppress
from threading import Thread

from aura_engine.base.api import LiquidsoapUtil as LU
from aura_engine.base.config import AuraConfig
from aura_engine.base.exceptions import (
    InvalidChannelException,
    LoadSourceException,
    LQConnectionError,
    LQStreamException,
)
from aura_engine.base.utils import DotDict
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.channels import (
    Channel,
    ChannelResolver,
    ChannelRouter,
    ChannelType,
    EntryPlayState,
    LiquidsoapResponse,
    ResourceType,
    TransitionType,
)
from aura_engine.client.connector import PlayerConnector
from aura_engine.control import EngineControlInterface
from aura_engine.events import EngineEventDispatcher
from aura_engine.mixer import Mixer, MixerType
from aura_engine.resources import ResourceClass, ResourceUtil


class Engine:
    """
    The Engine.
    """

    instance = None
    engine_time_offset = 0.0
    logger = None
    eci = None
    channels = None
    channel_router = None
    scheduler = None
    event_dispatcher = None
    plugins = None
    connector = None
    playout_state = None

    def __init__(self):
        """
        Initialize the engine.
        """
        if Engine.instance:
            raise Exception("Engine is already running!")
        Engine.instance = self
        self.logger = logging.getLogger("AuraEngine")
        self.config = AuraConfig.config()
        Engine.engine_time_offset = float(self.config.get("engine_latency_offset"))
        self.plugins = dict()
        self.channel_router = ChannelRouter(self.config, self.logger)
        self.start()

    def start(self):
        """
        Start the engine.

        Called when the connection to the sound-system implementation has been established.
        """
        self.event_dispatcher = EngineEventDispatcher(self)
        self.eci = EngineControlInterface(self, self.event_dispatcher)
        self.connector = PlayerConnector(self.event_dispatcher)
        self.event_dispatcher.on_initialized()

        while not self.is_connected():
            self.logger.info(SU.yellow("Waiting for Liquidsoap to be running ..."))
            time.sleep(2)
        self.logger.info(SU.green("Engine Core ------[ connected ]-------- Liquidsoap"))
        result = self.engine_update_config()
        if result == "OK":
            self.logger.info(SU.green("Playout config successfully updated"))
        else:
            self.logger.info(SU.red("Error while updating playout config"))

        self.player = Player(self.connector, self.event_dispatcher)
        self.event_dispatcher.on_boot()

        self.logger.info(EngineSplash.splash_screen(self.config))
        self.event_dispatcher.on_ready()

    #
    #   Basic Methods
    #

    def is_connected(self):
        """
        Check if there's a valid connection to Liquidsoap.
        """
        has_connection = False
        try:
            self.uptime()
            has_connection = True
        except LQConnectionError:
            self.logger.info("Liquidsoap is not running so far")
        except Exception as e:
            self.logger.error("Cannot check if Liquidsoap is running. Reason: " + str(e))

        return has_connection

    def update_playout_state(self):
        """
        Retrieve the state of all inputs and outputs.
        """
        state = self.connector.send_lqc_command("engine", "state")
        state = DotDict(LU.json_to_dict(state))

        # Initialize state
        if not self.playout_state:
            if state.is_fallback:
                self.logger.info(SU.yellow("Set initial playout state to FALLBACK"))
                timeslot = self.scheduler.programme.get_current_timeslot()
                fallback_show_name = self.config.get("fallback_show_name")
                self.event_dispatcher.on_fallback_active(timeslot, fallback_show_name)
            else:
                self.logger.info(SU.green("Set initial playout state"))
        elif state and self.playout_state.is_fallback != state.is_fallback:
            if state.is_fallback:
                self.logger.info(SU.yellow("Playout turned into FALLBACK state"))
                timeslot = self.scheduler.programme.get_current_timeslot()
                fallback_show_name = self.config.get("fallback_show_name")
                self.event_dispatcher.on_fallback_active(timeslot, fallback_show_name)
            else:
                self.logger.info(SU.green("Playout turned back into normal state"))

        self.playout_state = state
        return self.playout_state

    def get_playout_state(self):
        """
        Retrieve the state of Engine Core.
        """
        return self.playout_state

    def engine_update_config(self):
        """
        Update the config of playout with the current values.
        """
        playout_config = {
            "fallback_show_id": int(self.config.get("fallback_show_id")),
            "fallback_show_name": self.config.get("fallback_show_name"),
        }
        json_config = json.dumps(playout_config, ensure_ascii=False)
        res = self.connector.send_lqc_command("engine", "update_config", json_config)
        return res

    def init_version(self) -> dict:
        """
        Get the versions of Engine components and store in configuration.

        Returns:
            dict: {
                "control": Engine Control version
                "core": Engine Core version
                "liquidsoap": Liquidsoap version
            }

        """
        ctrl_version = None
        with open(os.path.join("", "VERSION")) as version_file:
            ctrl_version = version_file.read().strip()

        # FIXME Should be in one call
        core_version = self.connector.send_lqc_command("engine", "version")
        liq_version = self.connector.send_lqc_command("version", "")
        liq_version = liq_version.split(" ")[1]

        self.config.set("version_control", ctrl_version)
        self.config.set("version_core", core_version)
        self.config.set("version_liquidsoap", liq_version)

    def uptime(self):
        """
        Retrieve the uptime of Liquidsoap.
        """
        self.connector.enable_transaction()
        data = self.connector.send_lqc_command("uptime", "")
        self.connector.disable_transaction()
        return data

    @staticmethod
    def engine_time():
        """
        Get the engine perspective on time.

        Liquidsoap is slow in executing commands, therefore it's needed to schedule
        actions by (n) seconds in advance, as defined in the configuration file by
        the property `engine_latency_offset`. it's important to note that this method
        requires the class variable `EngineUtil.engine_time_offset` to be set on
        Engine initialization.

        Returns:
            (Integer):  the Unix epoch timestamp including the offset

        """
        return SU.timestamp() + Engine.engine_time_offset

    @staticmethod
    def get_instance():
        """
        Return the one and only engine.
        """
        return Engine.instance

    def terminate(self):
        """
        Terminate the engine and all related processes.
        """
        if self.scheduler:
            self.scheduler.terminate()
        if self.eci:
            self.eci.terminate()


#
#   PLAYER
#


class Player:
    """
    Engine Player.
    """

    config = None
    logger = None
    connector = None
    channels = None
    channel_router = None
    event_dispatcher = None
    mixer = None

    def __init__(self, connector, event_dispatcher):
        """
        Initialize the player.

        Args:
            connector (Connector): Connector for the playout
            event_dispatcher (EventDispather): Dispatcher for issuing events

        """
        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")
        self.event_dispatcher = event_dispatcher
        self.connector = connector
        self.channel_router = ChannelRouter(self.config, self.logger)
        self.mixer = Mixer(self.config, MixerType.MAIN, self.connector)

    def preload(self, entry):
        """
        Pre-Load the entry. This is required before the actual `play(..)` can happen.

        Be aware when using this method in order to queue a very short entry this may
        result in situations with incorrect timing. In this case bundle multiple short
        entries as one queue using `preload_playlist(self, entries)`.

        It's important to note, that his method is blocking until loading has finished. If this
        method is called asynchronously, the progress on the preloading state can be looked up in
        `entry.state`.

        Args:
            entry (Entry): An array holding filesystem entries

        """
        entry.status = EntryPlayState.LOADING
        self.logger.info("Loading entry '%s'" % entry)
        is_ready = False

        def set_metadata():
            track_meta = ResourceUtil.generate_track_metadata(entry, True)
            json_meta = json.dumps(track_meta, ensure_ascii=False)
            res = self.connector.send_lqc_command(entry.channel, "set_track_metadata", json_meta)
            self.logger.info(f"Response for '{entry.channel}.set_track_metadata': {res}")
            if res not in LiquidsoapResponse.SUCCESS.value:
                msg = f"Error while setting metadata on {entry.channel} to:\n{json_meta}"
                self.logger.error(SU.red(msg))

        # LIVE
        if entry.get_content_type() in ResourceClass.LIVE.types:
            entry.channel = ChannelResolver.live_channel_for_resource(entry.source)
            if entry.channel is None:
                self.logger.critical(SU.red("Invalid live channel '{entry.source}' requested!"))
            entry.previous_channel = None
            set_metadata()
            is_ready = True
        else:
            # Store channels for non-live entries
            channel_type = self.channel_router.type_for_resource(entry.get_content_type())
            entry.previous_channel, entry.channel = self.channel_router.channel_swap(channel_type)
            msg = f"Stored channel '{entry.channel}' in entry (prev: '{entry.previous_channel}')"
            self.logger.info(SU.pink(msg))

            # QUEUE
            if entry.get_content_type() in ResourceClass.FILE.types:
                metadata = ResourceUtil.generate_track_metadata(entry)
                is_ready = self.queue_push(entry.channel, entry.source, metadata)

            # STREAM
            elif entry.get_content_type() in ResourceClass.STREAM.types:
                is_ready = self.stream_load_entry(entry)

        if is_ready:
            set_metadata()
            entry.status = EntryPlayState.READY

        self.event_dispatcher.on_queue([entry])

    def preload_group(self, entries, channel_type=ChannelType.QUEUE):
        """
        Pre-Load multiple filesystem entries at once.

        This call is required before the actual `play(..)` can happen. Due to their nature,
        non-filesystem entries cannot be queued using this method. In this case use
        `preload(self, entry)` instead. This method also allows queuing of very short files, such
        as jingles.

        It's important to note, that his method is blocking until loading has finished. If this
        method is called asynchronously, the progress on the preloading state can be looked up in
        `entry.state`.

        Args:
            entries ([Entry]): An array holding filesystem entries
            channel_type (ChannelType): The type of channel where it should be queued (optional)

        """
        channels = None

        # Validate entry type
        for entry in entries:
            if entry.get_content_type() != ResourceType.FILE:
                raise InvalidChannelException

        # Determine channel
        channels = self.channel_router.channel_swap(channel_type)

        # Queue entries
        for entry in entries:
            entry.status = EntryPlayState.LOADING
            self.logger.info("Loading entry '%s'", entry)

            # Choose and save the input channel
            entry.previous_channel, entry.channel = channels
            metadata = ResourceUtil.generate_track_metadata(entry)
            if self.queue_push(entry.channel, entry.source, metadata):
                entry.status = EntryPlayState.READY

        self.event_dispatcher.on_queue(entries)
        return channels

    def play(self, entry, transition):
        """
        Play a new `Entry`.

        In case of a new timeslot (or some intended, immediate transition), a clean channel is
        selected and transitions between old and new channel is performed.

        This method expects that the entry is pre-loaded using `preload(..)` or
        `preload_group(self, entries)` before being played. In case the pre-loading has happened
        for a group of entries, only the first entry of the group needs to be passed.

        Args:
            entry (PlaylistEntry): The audio source to be played
            transition (TransitionType): The type of transition to use e.g. fade-in or instant
                volume level.
            queue (Boolean): If `True` the entry is queued if the `ChannelType` does allow so;
                otherwise a new channel of the same type is activated

        """
        with suppress(LQConnectionError):

            channel_type = self.channel_router.type_of_channel(entry.channel)
            mixer = self.mixer

            # Instant activation or fade-in
            self.connector.enable_transaction()
            if transition == TransitionType.FADE:
                mixer.channel_select(entry.channel.value, True)
                mixer.fade_in(entry.channel, entry.volume)
                msg = f"Select channel '{entry.channel}' with {transition}"
                self.logger.info(SU.pink(msg))
            else:
                mixer.channel_activate(entry.channel.value, True)
                msg = f"Activate channel '{entry.channel}'"
                self.logger.info(SU.pink(msg))
            self.connector.disable_transaction()

            # Update active channel for the current channel type
            self.channel_router.set_active(channel_type, entry.channel)
            prev_channel = entry.previous_channel
            if prev_channel in ChannelType.QUEUE.channels:
                self.queue_clear(prev_channel)
            self.event_dispatcher.on_play(entry)

            # Store most recent channel to timeslot, providing knowledge about
            # which channel should be faded out at the end of the timeslot
            timeslot = entry.playlist.timeslot
            timeslot.latest_channel = entry.channel
            msg = f"Stored recent entry's channel '{entry.channel}' to timeslot"
            self.logger.info(SU.pink(msg))

    def stop(self, channel: Channel, transition: TransitionType):
        """
        Stop the currently playing channel.

        Args:
            channel (Channel): The channel to stop playing or fade-out
            transition (TransitionType): The type of transition to use e.g. fade-out
        """
        with suppress(LQConnectionError):
            self.connector.enable_transaction()
            if not channel:
                self.logger.warn(SU.red("Cannot stop, no channel passed"))
                return

            if transition == TransitionType.FADE:
                self.mixer.fade_out(channel)
            else:
                self.mixer.channel_volume(channel, 0)

            self.logger.info(SU.pink(f"Mute channel '{channel}' with {transition}"))
            self.connector.disable_transaction()
            self.event_dispatcher.on_stop(channel)

    #
    #   Channel Type - Stream
    #

    def stream_load_entry(self, entry):
        """
        Load the given stream entry and updates the entries's status codes.

        Args:
            entry (Entry): The entry to be pre-loaded

        Returns:
            (Boolean):  `True` if successful

        """
        self.stream_load(entry.channel, entry.source)
        time.sleep(1)

        retry_delay = self.config.get("input_stream_retry_delay")
        max_retries = self.config.get("input_stream_max_retries")
        retries = 0

        while not self.stream_is_ready(entry.channel, entry.source):
            self.logger.info("Loading Stream ...")
            if retries >= max_retries:
                raise LoadSourceException(
                    "Could not connect to stream while waiting for %s seconds!"
                    % str(retries * retry_delay)
                )
            time.sleep(retry_delay)
            retries += 1

        return True

    def stream_load(self, channel, url):
        """
        Preload the stream URL on the given channel.

        Note this method is blocking some serious amount of time; hence it is worth being called
        asynchronously.

        Args:
            channel (Channel): The stream channel
            url (String): The stream URL

        Returns:
            (Boolean): `True` if successful

        """
        result = None

        self.connector.enable_transaction()
        result = self.connector.send_lqc_command(channel, "stream_stop")

        if result not in LiquidsoapResponse.SUCCESS.value:
            self.logger.error("%s.stop result: %s" % (channel, result))
            raise LQStreamException("Error while stopping stream!")

        result = self.connector.send_lqc_command(channel, "stream_set_url", url)

        if result not in LiquidsoapResponse.SUCCESS.value:
            self.logger.error("%s.set_url result: %s" % (channel, result))
            raise LQStreamException("Error while setting stream URL!")

        # Liquidsoap ignores commands sent without a certain timeout
        time.sleep(2)

        result = self.connector.send_lqc_command(channel, "stream_start")
        self.logger.info("%s.start result: %s" % (channel, result))

        self.connector.disable_transaction()
        return result

    def stream_is_ready(self, channel, url):
        """
        Check if the stream on the given channel is ready to play.

        Note this method is blocking some serious amount of time even when successful; hence it is
        worth being called asynchronously.

        Args:
            channel (Channel): The stream channel
            url (String): The stream URL

        Returns:
            (Boolean): `True` if successful

        """
        result = None

        self.connector.enable_transaction()

        result = self.connector.send_lqc_command(channel, "stream_status")
        self.logger.info("%s.status result: %s" % (channel, result))

        if not result.startswith(LiquidsoapResponse.STREAM_STATUS_CONNECTED.value):
            return False

        lqs_url = result.split(" ")[1]
        if not url == lqs_url:
            self.logger.error(
                "Wrong URL '%s' set for channel '%s', expected: '%s'." % (lqs_url, channel, url)
            )
            return False

        self.connector.disable_transaction()

        stream_buffer = self.config.get("input_stream_buffer")
        self.logger.info(
            "Ready to play stream, but wait %s seconds until the buffer is filled..."
            % str(stream_buffer)
        )
        time.sleep(round(float(stream_buffer)))
        return True

    #
    #   Channel Type - Queue
    #

    def queue_push(self, channel, source, metadata):
        """
        Add an filesystem URI to the given `ChannelType.QUEUE` channel.

        Args:
            channel (Channel): The channel to push the file to
            source (String): The URI of the file
            metadata (dict): additional meta data to be wrapped with the URI

        Returns:
            (Boolean): `True` if successful

        """
        if (
            channel not in ChannelType.QUEUE.channels
            and channel not in ChannelType.FALLBACK_QUEUE.channels
        ):
            raise InvalidChannelException

        self.connector.enable_transaction()
        audio_store = self.config.abs_audio_store_path()
        extension = self.config.get("audio_source_extension")
        filepath = ResourceUtil.source_to_filepath(audio_store, source, extension)
        self.logger.info(SU.pink(f"{channel}.queue_push('{filepath}')"))
        if metadata:
            filepath = ResourceUtil.lqs_annotate(filepath, metadata)
        result = self.connector.send_lqc_command(channel, "queue_push", filepath)
        self.logger.info("%s.queue_push result: %s" % (channel, result))
        self.connector.disable_transaction()

        # If successful, Liquidsoap returns a resource ID of the queued track
        resource_id = -1
        try:
            resource_id = int(result)
        except ValueError:
            self.logger.error(SU.red("Got an invalid resource ID: '%s'" % result))
            return False

        return resource_id >= 0

    def queue_seek(self, channel, seconds_to_seek):
        """
        Pre-roll the player of the given `ChannelType.QUEUE` channel by (n) seconds.

        Args:
            channel (Channel): The channel to push the file to
            seconds_to_seeks (Float): The seconds to skip

        Returns:
            (String): Liquidsoap response

        """
        if (
            channel not in ChannelType.QUEUE.channels
            and channel not in ChannelType.FALLBACK_QUEUE.channels
        ):
            raise InvalidChannelException

        self.connector.enable_transaction()
        result = self.connector.send_lqc_command(channel, "queue_seek", str(seconds_to_seek))
        self.logger.info("%s.seek result: %s" % (channel, result))
        self.connector.disable_transaction()

        return result

    def queue_clear(self, channel: Channel):
        """
        Clear any tracks left in the queue.

        Dear queue channels, please leave the room as you would like to find it:
        Every queue needs to be empty, before it can receive new items. Otherwise old tracks
        continue to play, when such 'dirty' channel is selected.

        The channel is cleared asynchronously.

        Args:
            channel (Channel): The channel to clear

        """
        if not self.channel_router.is_any_queue(channel):
            raise InvalidChannelException

        def clean_up():
            # Wait some moments, if there is some long fade-out. Note, this also
            # means, this channel should not be used for at least some seconds
            # (including clearing time).
            clear_timeout = 10
            self.logger.info(f"Clearing channel '{channel}' in {clear_timeout} seconds")
            time.sleep(clear_timeout)

            # Deactivate channel
            self.connector.enable_transaction()
            response = self.mixer.channel_activate(channel.value, False)
            self.connector.disable_transaction()
            msg = f"Deactivate channel '{channel}' with message '{response}'"
            self.logger.info(SU.pink(msg))

            # Remove all items from queue
            self.connector.enable_transaction()
            result = self.connector.send_lqc_command(channel, "queue_clear")
            self.connector.disable_transaction()
            msg = f"Clear queue channel '{channel}' with result: {result}"
            self.logger.info(SU.pink(msg))

            # FIXME Workaround to get clearing somewhat working (See engine-core#16)
            self.mixer.channel_activate(channel.value, True)
            self.mixer.channel_activate(channel.value, False)

        Thread(target=clean_up).start()

    #
    #   Channel Type - Playlist
    #

    def playlist_set_uri(self, channel, playlist_uri):
        """
        Set the URI of a playlist.

        Args:
            channel (Channel): The channel to push the file to
            playlist_uri (String): The path to the playlist

        Returns:
            (String): Liquidsoap response

        """
        self.logger.info(SU.pink("Setting URI of playlist '%s' to '%s'" % (channel, playlist_uri)))

        self.connector.enable_transaction()
        result = self.connector.send_lqc_command(channel, "playlist_uri_set", playlist_uri)
        self.logger.info("%s.playlist_uri result: %s" % (channel, result))
        self.connector.disable_transaction()

        return result

    def playlist_clear_uri(self, channel):
        """
        Clear the URI of a playlist.

        Args:
            channel (Channel): The channel to push the file to

        Returns:
            (String): Liquidsoap response

        """
        self.logger.info(SU.pink("Clearing URI of playlist '%s'" % (channel)))

        self.connector.enable_transaction()
        result = self.connector.send_lqc_command(channel, "playlist_uri_clear")
        self.logger.info("%s.playlist_uri_clear result: %s" % (channel, result))
        self.connector.disable_transaction()

        return result


class EngineSplash:
    """Print the splash and version information on boot."""

    @staticmethod
    def splash_screen(config):
        """
        Print the engine logo and version info.
        """
        version = config.get("version_control")
        core_version = config.get("version_core")
        liq_version = config.get("version_liquidsoap")
        return f"""\n
             █████╗ ██╗   ██╗██████╗  █████╗     ███████╗███╗   ██╗ ██████╗ ██╗███╗   ██╗███████╗
            ██╔══██╗██║   ██║██╔══██╗██╔══██╗    ██╔════╝████╗  ██║██╔════╝ ██║████╗  ██║██╔════╝
            ███████║██║   ██║██████╔╝███████║    █████╗  ██╔██╗ ██║██║  ███╗██║██╔██╗ ██║█████╗
            ██╔══██║██║   ██║██╔══██╗██╔══██║    ██╔══╝  ██║╚██╗██║██║   ██║██║██║╚██╗██║██╔══╝
            ██║  ██║╚██████╔╝██║  ██║██║  ██║    ███████╗██║ ╚████║╚██████╔╝██║██║ ╚████║███████╗
            ╚═╝  ╚═╝ ╚═════╝ ╚═╝  ╚═╝╚═╝  ╚═╝    ╚══════╝╚═╝  ╚═══╝ ╚═════╝ ╚═╝╚═╝  ╚═══╝╚══════╝
            control v{version}, core v{core_version}, liquidsoap v{liq_version} - Ready to play!
        \n"""
